@@ -93,7 +93,7 @@ class AISMessages(object):
             msg_dict = msg.asdict(enum_as_int=enum_as_int, ignore_spare=ignore_spare)
             
             # Extract tagblock of first sentence if present, and add tagblock information in result
-            tag_block_str, _ = self.split_sentence_to_tab_block_and_message(sentences[0])
+            tag_block_str, _ = self.split_sentence_to_tag_block_and_message(sentences[0])
             if tag_block_str is None:
                 tag_block_dict = None
             else:
@@ -283,7 +283,7 @@ class AISMessages(object):
         
         return res
     
-    def split_sentence_to_tab_block_and_message(self, ais_sentence):
+    def split_sentence_to_tag_block_and_message(self, ais_sentence):
         if isinstance(ais_sentence, str):
             split_char = '\\'
         elif isinstance(ais_sentence, bytes):
