@@ -299,7 +299,7 @@ class RMQClient(DeleteableObject):
             logger.debug(f"[{self.name}] Start processing data events...")
         self.__is_processing_data_events = True
         
-        self.__process_data_events_thread = LoopFunctionThreaded(self.connection.process_data_events, kwargs={'time_limit':1}, register_thread=True, delay_before_run_sec=None)
+        self.__process_data_events_thread = LoopFunctionThreaded(self.connection.process_data_events, kwargs={'time_limit':1}, register_thread=True, delay_before_run_sec=None, delay_between_run_sec=0.01)
         self.__process_data_events_thread.start()
 
     def stop_process_data_events(self):
