@@ -10,13 +10,15 @@
 # The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
 #################################################
 
-# from flask.views import MethodView
-# from holado.common.context.session_context import SessionContext
+from flask.views import MethodView
 # import logging
-#
+
 # logger = logging.getLogger(__name__)
-#
-#
-# def _get_session_context():
-#     return SessionContext.instance()
+
+
+class HealthView(MethodView):
+    
+    def get(self):
+        return {'status': 'healthy'}
+
 
